
#include "arpc.h"
#include "sfscrypt.h"
#include "rxx.h"
#include "sfskeymisc.h"

sfscrypt sfscrpt;
static rxx comma (",");
static rxx hexrx ("0x[\\da-fA-F]+");

ptr<sfspriv>
sfscrypt::alloc (const sfs_privkey2_clear &pk, u_char o) const
{
  sfsca *c = xttab[pk.type];
  if (!c) return NULL;
  return c->alloc (pk, o);
}

ptr<sfspriv>
sfscrypt::alloc (const str &ske, str *kn, str *pwd, u_int *cost, 
		 u_char o) const
{
  vec<str> kv;
  if (split (&kv, comma, ske, 2, true) != 2)
    return NULL;
  sfsca *c = strtab[kv[0]];
  if (!c) return NULL;
  return c->alloc (ske, kn, pwd, cost, o);
}


ptr<sfspub>
sfscrypt::alloc (const sfs_pubkey2 &pk, u_char o) const
{
  sfsca *c = xttab[pk.type];
  if (!c) return NULL;
  return c->alloc (pk, o);
}

ptr<sfspub>
sfscrypt::alloc (const str &s, u_char o) const
{
  vec<str> kv;
  if (split (&kv, comma, s, 2, true) != 2)
    return NULL;
  sfsca *c = strtab[kv[0]];
  if (!c) return NULL;
  return c->alloc (kv[1], o);
}

ptr<sfspriv>
sfscrypt::alloc (sfs_keytype xt, const str &epk, const eksblowfish *eksb,
		 u_char o) const
{
  sfsca *c = xttab[xt];
  if (!c) return NULL;
  return c->alloc (epk, eksb, o);
}

ptr<sfspriv>
sfscrypt::alloc (sfs_keytype xt, u_int nbits, u_char o) const
{
  sfsca *c = xttab[xt];
  if (!c) return NULL;
  return c->alloc (nbits, o);
}

ptr<sfspub>
sfscrypt::alloc (const sfs_pubkey &pk, u_char o) const 
{
  sfsca *c = xttab[SFS_RABIN];
  if (!c) return NULL;
  return c->alloc (pk, o);
}

ptr<sfspriv>
sfscrypt::alloc_priv (const str &sk, u_char o) const
{
  vec<str> kv;
  if (split (&kv, comma, sk, 2, true) != 2)
    return NULL;
  sfsca *c = strtab[kv[0]];
  if (!c) return NULL;
  return c->alloc_priv (sk, o);
}

ptr<sfspub>
sfs_rabin_alloc::alloc (const sfs_pubkey &k, u_char o) const
{
  ptr<rabin_pub> rk = New refcounted<rabin_pub> (k);
  return New refcounted<sfs_rabin_pub, vbase> (rk, o);
}


ptr<sfspriv>
sfs_rabin_alloc::alloc (u_int nbits, u_char o) const
{
  ptr<rabin_priv> k = New refcounted<rabin_priv> (rabin_keygen (nbits));
  return New refcounted<sfs_rabin_priv, vbase> (k, o);
}

ptr<sfspriv>
sfs_rabin_alloc::alloc_priv (const str &sk, u_char o) const
{
  ptr<rabin_priv> k = import_rabin_priv (sk, NULL);
  if (!k) return NULL;
  return New refcounted<sfs_rabin_priv, vbase> (k, o);
}

ptr<sfspriv>
sfs_rabin_alloc::alloc (const str &epk, const eksblowfish *eksb, 
			u_char o) const
{
  ptr<rabin_priv> k = import_rabin_decrypt_sec (epk, eksb, false);
  if (!k) return NULL;
  return New refcounted<sfs_rabin_priv, vbase> (k, o);
}


ptr<sfspriv>
sfs_rabin_alloc::alloc (const sfs_privkey2_clear &k, u_char o) const
{
  assert (k.type == ktype);
  ptr<rabin_priv> rpk = New refcounted<rabin_priv> (k.rabin->p, k.rabin->q);
  if (!rpk) return NULL;
  return New refcounted<sfs_rabin_priv, vbase> (rpk, o);
}

ptr<sfspub>
sfs_rabin_alloc::alloc (const sfs_pubkey2 &k, u_char o) const
{
  assert (k.type == ktype) ;
  ptr<rabin_pub> rpk = New refcounted<rabin_pub> (*(k.rabin));
  if (!rpk) return NULL;
  return New refcounted<sfs_rabin_pub, vbase> (rpk, o);
}

ptr<sfspriv>
sfs_rabin_alloc::alloc (str ske, str *kn, str *pwd, u_int *cost, 
			u_char o) const
{
  assert (pwd && cost && kn);
  if (import_rabin_priv_need_pwd (ske, cost))
    *pwd = getpwd ("Passphrase for " << *kn << ": ");
  ptr<rabin_priv> rsk = import_rabin_priv (ske, *pwd, kn);
  if (!rsk) {
    warn << "Bad passphrase." ;
    return NULL;
  }
  return New refcounted<sfs_rabin_priv, vbase> (rsk, o);
}

ptr<sfspub>
sfs_rabin_alloc::alloc (const str &pk, u_char o) const
{
  if (!hexrx.match (pk)) {
    warn << "Malformed Rabin public key given\n";
    return NULL;
  }
  ptr<rabin_pub> rpk = New refcounted<rabin_pub> (bigint (pk.cstr ()));
  if (!rpk)
    return NULL;
  return New refcounted<sfs_rabin_pub, vbase> (rpk, o);
}

bool
sfscrypt::verify (const sfs_pubkey2 &pk, const sfs_sig2 &sig,
		   const str &msg) const
{
  ptr<sfspub> p = alloc (pk, SFS_VERIFY);
  if (!p) {
    warn << "Could not use public key to verify.\n";
    return false;
  }
  return p->verify (sig, msg);
}

bool
sfspriv::decrypt (const sfs_ctext2 &ct, str *msg)
{
  if (!load_ctext (ct))
    return false;
  return decrypt (msg);
}

bool
sfspriv::decrypt (const sfs_ctext &ct, str *msg)
{
  if (!load_ctext (ct))
    return false;
  return decrypt (msg);
}

bool
sfspub::operator== (const sfspub &p) const
{
  sfs_pubkey2 pk;
  return (export_pubkey (&pk) && p == pk);
}

bool
sfspub::operator== (const str &b) const
{
  strbuf f;
  if (!export_pubkey (f))
    return false;
  return (str (f) == b);
}

bool 
sfspub::verify_init (const sfs_sig2 &sig, const str &msg) const
{
  size_t nbits = get_pubkey_nbits ();
  if (nbits < sfs_minpubkeysize) {
    if (nbits)
      warn ("%d-bit pub key too short\n", nbits);
    return false;
  }
  if (sig.type != ktype) {
    warn ("pub key type %d does not match sig type %d\n", sig.type, ktype);
    return false;
  }
  return true;
}

bool 
sfs_rabin_priv::sign (sfs_sig2 *sig, const str &msg)
{
  assert (get_opt (SFS_SIGN));
  sig->set_type (SFS_RABIN);
  if ((*sig->rabin = privk->sign (msg)) != 0) return true;
  else return false;
}

bool 
sfs_rabin_pub::verify (const sfs_sig2 &sig, const str &msg)
{
  assert (get_opt (SFS_VERIFY));
  if (!verify_init (sig, msg)) return false;
  bigint n = *(sig.rabin);
  return pubk->verify (msg, *(sig.rabin));
}

bool
sfs_rabin_pub::verify_r (const bigint &n, size_t len, str &msg)
{
  assert (get_opt (SFS_VERIFY));
  msg = pubk->verify_r (n, len);
  if (msg) return true;
  else return false;
}

bool
sfs_rabin_priv::sign_r (sfs_sig *sig, const str &msg)
{
  assert (get_opt (SFS_SIGN));
  if (!(*sig = privk->sign_r (msg))) return false;
  return true;
}

bool
sfs_rabin_priv::export_privkey (sfs_privkey2_clear *k) const
{
  k->set_type (ktype);
  k->rabin->p = privk->p;
  k->rabin->q = privk->q;
  return true;
}

bool
sfs_rabin_priv::export_privkey (str *s, const str &kn, 
				str pwd, u_int cost) const
{
  return (*s = export_rabin_priv (*privk, pwd, kn, cost));
}

bool
sfs_rabin_priv::export_privkey (str *s, const eksblowfish *eksb) const
{
  return (*s = export_rabin_encrypt_sec (*privk, eksb, false));
}

bool
sfs_rabin_pub::export_pubkey (sfs_pubkey2 *k) const
{
  k->set_type (ktype);
  *k->rabin = pubk->n;
  return true;
}

bool
sfs_rabin_pub::operator== (const sfs_pubkey2 &pk) const
{
  if (pk.type != ktype) return false;
  return (pubk->n == *pk.rabin);
}


bool
sfs_rabin_pub::operator== (const str &pk) const
{
  vec<str> kv;
  if (split (&kv, comma, pk, 2, true) != 2)
    return false;
  if (kv[0] != "rabin") return false;
  if (pubk->n != bigint (kv[1])) return false;
  return true;
}

size_t
sfs_rabin_pub::get_pubkey_nbits () const
{
  return pubk->n.nbits ();
}

bool
sfs_rabin_pub::export_pubkey (strbuf &b) const
{
  b << "rabin,0x" << pubk->n.getstr (16);
  return true;
}

bool
sfs_rabin_pub::encrypt (sfs_ctext2 *ct, const str &msg)
{
  assert (get_opt (SFS_ENCRYPT));
  ct->set_type (ktype);
  if (!(*(ct->rabin) = pubk->encrypt (msg))) return false;
  return true;
}

bool
sfs_rabin_priv::decrypt (str *msg)
{
  assert (get_opt (SFS_DECRYPT));
  if (!ctext) return false;
  *msg = privk->decrypt (ctext, sizeof (sfs_kmsg));
  if (!msg || msg->len () < 0) return false;
  return true;
}

bool 
sfs_rabin_priv::load_ctext (const sfs_ctext &ct) 
{
  ctext = ct;
  return true;
}

bool
sfs_rabin_priv::load_ctext (const sfs_ctext2 &ct)
{
  if (ct.type != ktype) return false;
  if (!ct.rabin) return false;
  ctext = *ct.rabin;
  return true;
}


bool
sfs_rabin_pub::encrypt (sfs_ctext *n, const str &msg) 
{
  assert (get_opt (SFS_ENCRYPT));
  if (!(*n = pubk->encrypt (msg))) return false;
  return true;
}

bool
sfs_rabin_pub::export_pubkey (sfs_pubkey *k) const 
{
  if (!(*k = pubk->n)) return false;
  return true;
}

bool 
sfs_rabin_priv::get_privkey_hash (u_int8_t *buf, const sfs_hash &hostid) const
{
  str p;
  sha1ctx sc;
  p = str2wstr (privk->p.getraw ());
  sc.update (p, p.len ());
  sc.update (hostid.base (), hostid.size ());
  p = str2wstr (privk->q.getraw ());
  sc.update (p, p.len ());
  sc.final (buf);
  return true;
}

