/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef SNAPSHOT_BUILD_H
#define SNAPSHOT_BUILD_H

#include "vec.h"
#include "ihash.h"
#include "stdinc.h"
#include "vervec.h"
#include "ns.h"
#include "cfs.h"
#include "snapshot.h"
#include "cfsw_record.h"

class snapshot;
class replica_impl;

typedef callback<void, ptr<snapshot>, cfsrw_status>::ptr cbs_t;

//
// builds one snapshot from a base snapshot and a set of deltas
//
class bob : public virtual refcount
{
  replica_impl *_replica;
  bigint projid;
  ptr<snapshot> sobj;
  vec<deltap> deltas;
  crec cr;
  commit_info commit;
  cbs_t scb;

  // storing snapshots
  int outstanding_inserts;
  int insert_failures;
  int inserts;
  int fmap_block_inserts;
  int fmap_block_insert_failures;
  int nchecks;
  vec <bigint> to_store;

  void store_tree ();
  void store_new (snode *);
  bigint store (bigint);
  bigint store_inode (bigint, cfs_inode *);
  bigint store_file (bigint, cfs_inode *);
  bigint store_directory (bigint, cfs_inode *);
  bigint store_one_fmap_block (str);
  
  void serialize_and_store ();

  void insert_cb (bool, bigint, ns_stat, bigint);
  bigint insert (char *, size_t);
  bigint insert_and_addtofmap (char *, size_t, bigint);
  void insert_snapshot ();
  void insert_fmapblock_cb (ns_stat, bigint);
 
  // building a new snapshot
  struct xr { deltap record; };
  typedef ptr<xr> xrp;
  
  bool storing_tree;
  int outstanding_treeops;
  int treeop_failures;
  cfsrw_status error;
  int history_update_rpcs;
  unsigned ndeltas;

  void next ();
  void build_tree ();

  bigint write_chain (cfs_vchain &c);
  void insert_history (cfs_history *, cfs_version &v);
  int incr_revision
    (cfs_attr *attr, cfs_history *hist, xrp &p, bool mode = false);
  
  void create_inode (xrp, cfs_inode *);
  
  void do_link (xrp);
  void do_link_got_dir (xrp, snode*);
  void do_link_got_node (xrp, snode*, snode*);
  void do_link_got_dirblk (xrp, snode*, snode*, uint32, ptr<cfs_data>, snode*);
  void add_dirent (xrp, cfs_inode_reg*, ptr<cfs_data>, snode*);

  void do_write (xrp);
  void do_write_got_node (xrp, snode*);
  void do_symlink (xrp, cfs_inode_lnk *);
  void write_file (xrp, cfs_inode_reg *);
  
  void do_unlink (xrp, int);
  void do_unlink_got_dir (xrp, int, snode*);
  void do_unlink_got_dirblk (xrp, int, int);
  void do_unlink_fetch_dirblk_cb (callback<void,void>::ref cb, snode *);
  void remove_dirent (xrp, int, snode *, cfs_data *, uint32);
  void remove_dirent_got_fh (xrp, int, bigint, cfs_inode_reg *, snode *);
  void done_unlink (xrp, int, bool err = false);
  void unlink_check_dir_block (cfs_inode *inode, cbb cb,
                               uint32 blocknum, ptr<cfs_data> data);
  void unlink_check_dir_empty (cfs_inode *inode, cbb cb);
  void unlink_dir_empty (xrp tp, int phase, cfs_inode_reg *dnode, bool empty);
  void do_rename (xrp);
  
  void setattr (xrp);
  void setattr_got_node (xrp, snode*);

  void do_patch (xrp x);
  void do_patch_got_node (xrp x, snode *f);
  void do_patch_read_cb (str tfn, int fd, xrp x, cfsrw_status status);
  void do_patch_got_patch (str tfn, xrp x, ptr<ns_block> b);
  void do_patch_write_patch (str tfn, xrp x, cfsw_write_arg a,
                             cfsrw_status status);

  void check (unsigned i);
  void check_node (unsigned i, bigint fh, snode *n);

  void check_write_fetch (ptr<ns_block> b);
  void check_write_cb (bigint n, bool x);
  void iter_write_cb (bigint n, cfsrw_status status);
  
  static ptr<snapshot> new_snapshot (ptr<snapshot> prev_s);

public:
  bob (replica_impl *r, bigint va, ptr<snapshot> sobj,
       const vec<deltap>& deltas, crec cr, commit_info commit,
       cbs_t scb);
};

inline void
init_history (cfs_history *history)
{
  history->versions.setsize (0);
  history->chain.v.revision = 0;
}

inline void
init_attr (cfs_inode_reg *inode)
{
  inode->attr.size = 0;
  inode->attr.mode = 0;
  inode->attr.revision = 0;
}

inline void
init_attr (cfs_inode_lnk *inode)
{
  inode->attr.size = 0;
  inode->attr.mode = 0;
  inode->attr.revision = 0;
}

inline void
set_attr (cfs_inode_reg *inode, cfsw_sattr *ca)
{
  if (ca->valid_size)
    inode->attr.size = ca->size;
  if (ca->valid_mode)
    inode->attr.mode = ca->mode;
}

inline void
set_attr (cfs_inode_lnk *inode, cfsw_sattr *ca)
{
  if (ca->valid_size)
    inode->attr.size = ca->size;
  if (ca->valid_mode)
    inode->attr.mode = ca->mode;
}

inline void 
getrnd (bigint *n, uint len) 
{
  char xx[len];
  rnd.getbytes (xx, len);
  mpz_set_rawmag_be (n, xx, len);
}

#endif

