/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

/* Parts of this file were written by Frans Kaashoek and Kevin Fu. */

%#include "bigint.h"
%#include "sfs_prot.h"
%#include "vervec.h"

%#ifndef HAVE_RPC_TRAVERSE
%inline bool
%rpc_traverse (rpc_clear_t &, bigint &obj)
%{
%  obj = 0;
%  return true;
%}
%#endif

%class cfs_data;
%typedef callback<void, ptr<cfs_data> >::ref cfs_retrieve_t;
%extern void ns_retrieve_cfs (bigint n, cfs_retrieve_t cb);
%extern void ns_retrieve_cfs_readonly (bigint n, cfs_retrieve_t cb);

%// there are serious consequences to changing these values, you may
%// not be able to read previously inserted snapshot or file blocks
%// anymore!!!

const SN_BLKSIZE = 8100;
const SN_NFH = 200;
const CFS_FHSIZE = 20;
const CFS_NDIR = 128;
const CFS_NVH = 32;
const CFS_NIND = 200;
const CFS_VCN = 31;

enum cfs_ftype {
  CFSREG = 0,
  CFSDIR = 1,
  CFSLNK = 2
};

struct cfs_version {
  uint32 revision;
  uint32 timestamp;
  bigint logid;
  bigint commit;
  bigint saddr;
};

struct cfs_vchain {
  cfs_version v;
  uint32 ptrs[CFS_VCN];
  bigint paddr[CFS_VCN];
};

struct cfs_history {
  cfs_version versions<CFS_NVH>;
  cfs_vchain chain;
};

struct cfs_attr {
  uint64 size;
  uint32 mode;
  uint32 nlink;
  bigint commit;
  bigint mode_commit;
  uint32 revision;
};

struct cfs_inode_lnk {
  bigint handle;
  cfs_attr attr;
  nfspath3 dest;
  cfs_history history;
};

struct cfs_inode_reg {
  bigint handle;
  cfs_attr attr;
  bigint direct<CFS_NDIR>;
  bigint indirect;
  bigint double_indirect;
  bigint triple_indirect;
  cfs_history history;
};

union cfs_inode switch (cfs_ftype type) {
case CFSLNK:
  cfs_inode_lnk lnk;
default:
  cfs_inode_reg reg;
};

struct cfs_indirect {
  bigint handles<>;
};

struct cfs_dirent {
  bigint fh;
  string name<>;
  cfs_dirent *nextentry;
};

struct cfs_directory {
  uint32 count;
  cfs_dirent *entries;
  bool eof;
};

struct cfs_rpc_dirent {
  bigint fh;
  string name<>;
  uint64 cookie;
  cfs_rpc_dirent *nextentry;
};

struct cfs_dirlist { 
  cfs_rpc_dirent *entries;
  bool eof;
};

enum dtype {
   CFS_INODE   = 0,
   CFS_FILEBLK = 1, /* File data */
   CFS_DIRBLK  = 2, /* Directory data */
   CFS_INDIR   = 3  /* Indirect data pointer block */
};

union cfs_data switch (dtype type) {
 case CFS_INODE:
   cfs_inode inode;
 case CFS_FILEBLK:
   opaque data<>;
 case CFS_DIRBLK:
   cfs_directory dir;
 case CFS_INDIR:
   cfs_indirect indir;
 default:
   void;
};


