/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "pastdbfe.h"
#include "vervec.h"
#include "pack.h"
#include "snapshot.h"
#include "load.h"

extern ptr<pastdbfe> create_inmem_db ();

void
load_snapshot (bigint sn_addr, cbloadsn_t cb)
{
  ptr<snapshot_loader> s = New refcounted<snapshot_loader> (sn_addr, cb);
}

snapshot_loader::snapshot_loader (bigint sn_addr, cbloadsn_t cb)
  : scb (cb)
{
  sobj = New refcounted<snapshot> ();
  sobj->thisaddr = sn_addr;
  get_fmap_cb (nsobj->retrieve_local (sn_addr));
}

void 
snapshot_loader::get_fmap_cb (ptr<ns_block> blk) 
{
  if (!blk) {
    warn << "cannot retrieve snapshot block (1).  Snapshot addr: " 
	 << armor64A(sobj->thisaddr) << "\n";
    scb ((ptr<snapshot>)0, CFSRWERR_ERROR);
    return;
  }
  parse_fmap (true, str (blk->data, blk->len));
}

void 
snapshot_loader::get_more_fmap_cb (ptr<ns_block> blk) 
{
  if (!blk) {
    warn << "cannot retrieve snapshot block (2).  Snapshot addr: " 
	 << armor64A(sobj->thisaddr) << "\n";
    scb ((ptr<snapshot>)0, CFSRWERR_ERROR);
    return;
  }
  parse_fmap (false, str (blk->data, blk->len));
}

void
snapshot_loader::parse_fmap (bool first, str ins)
{
  xin in (ins);

  if (first) {
    str swver;
    if (!in.rstr (swver) || !sobj->thisinfo.parse (in)) {
      warn << "cannot parse snapshot\n";
      scb ((ptr<snapshot>)0, CFSRWERR_ERROR);
      return;
    }
    sobj->thisinfo.branch.vptr = sobj->thisaddr;
    if (!(sobj->fmap_db = create_inmem_db ())) {
      warn << "cannot create fmap database\n";
      scb ((ptr<snapshot>)0, CFSRWERR_ERROR);
      return;
    }
  }

  int n;
  int remain;
  bigint next;

  in.rint (n);
  in.rint (remain);
  in.rbigint (next);

  for (int i=0; i<n; i++) {
    bigint inum, chash;
    in.rbigint (inum);
    in.rbigint (chash);

    ptr<dbrec> k = bigint2dbrec (inum);
    ptr<dbrec> d = bigint2dbrec (chash);
    if (sobj->fmap_db->insert (k, d)) {
      scb ((ptr<snapshot>)0, CFSRWERR_ERROR);
      return;
    }
  }

  if (remain > 0)
    get_more_fmap_cb (nsobj->retrieve_local (next));
  else
    scb (sobj, CFSRW_OK);
}

