#include "loctabroute.h"
#include "chord_impl.h"
#include "route_loctab.h"
#include <blacklist.h>
#include <location.h>
#include <locationtable.h>
#include <configurator.h>


loctabroute::loctabroute (ref<chord> _chordnode, ref<rpc_manager> _rpcm,
			  ref<location> _l)
  : vnode_impl( _chordnode, _rpcm, _l )
{
  verify_nsucc();
}

loctabroute::loctabroute (str blacklist_dir, ref<chord> _chordnode, 
			  ref<rpc_manager> _rpcm, ref<location> _l)
  : vnode_impl( _chordnode, _rpcm, _l )
{
  verify_nsucc();
  blacklist::only().set_dir(blacklist_dir);
}


void
loctabroute::verify_nsucc() 
{
  // This should be inferred from the hostlist
  Configurator::only().get_int("chord.nsucc", nsucc); 
  assert( (unsigned) nsucc >= hostlist::only().size()-1 );
}

loctabroute::~loctabroute (void) {
}

ref<vnode>
loctabroute::produce_vnode
  (str blacklist_dir, ref<chord> _chordnode, ref<rpc_manager> _rpcm, ref<location> _l)
{
  return New refcounted<loctabroute> (blacklist_dir, _chordnode, _rpcm, _l);

}

ptr<route_iterator> 
loctabroute::produce_iterator (chordID xi) 
{
  return New refcounted<route_loctab> (mkref (this), xi, nsucc);
}

ptr<route_iterator> 
loctabroute::produce_iterator (chordID xi, const rpc_program &uc_prog,
			       int uc_procno, ptr<void> uc_args) 
{
  return New refcounted<route_loctab>
    (mkref (this), xi, uc_prog, uc_procno, uc_args, nsucc);
}
 
route_iterator *
loctabroute::produce_iterator_ptr (chordID xi) 
{
  return New route_loctab (mkref (this), xi, nsucc);
}

route_iterator *
loctabroute::produce_iterator_ptr (chordID xi, const rpc_program &uc_prog,
				   int uc_procno, ptr<void> uc_args) 
{
  return New route_loctab
    (mkref (this), xi, uc_prog, uc_procno, uc_args, nsucc);
}


ptr<route_iterator> 
loctabroute::produce_iterator (chordID xi, bigint projkey) 
{
  return New refcounted<route_loctab> (mkref (this), mkref(this), xi, projkey, nsucc);
}

ptr<route_iterator> 
loctabroute::produce_iterator (chordID xi, bigint projkey, const rpc_program &uc_prog,
			       int uc_procno, ptr<void> uc_args) 
{
  return New refcounted<route_loctab> (mkref (this), mkref(this), xi, projkey, uc_prog, 
				       uc_procno, uc_args, nsucc);
}
 
route_iterator *
loctabroute::produce_iterator_ptr (chordID xi, bigint projkey) 
{
  return New route_loctab (mkref (this), mkref(this), xi, projkey, nsucc);
}

route_iterator *
loctabroute::produce_iterator_ptr (chordID xi, bigint projkey, const rpc_program &uc_prog,
				   int uc_procno, ptr<void> uc_args) 
{
  return New route_loctab (mkref (this), mkref(this), xi, projkey, uc_prog, uc_procno, uc_args, nsucc);
}

void
loctabroute::find_succlist (const chordID &x, u_long m, cbroute_t cb,
			   ptr<chordID> guess)
{
  route_iterator *ri = produce_iterator_ptr (x);
  ri->first_hop (wrap (this, &loctabroute::find_succlist_hop_cb, cb, ri, m),
		 guess);
}

void
loctabroute::find_succlist (const chordID &x, const bigint projkey, u_long m, cbroute_t cb,
			   ptr<chordID> guess)
{
  route_iterator *ri = produce_iterator_ptr (x, projkey);
  ri->first_hop (wrap (this, &loctabroute::find_succlist_hop_cb, cb, ri, m),
		 guess);
}

void
loctabroute::find_successor (const chordID &x, cbroute_t cb)
{
  nfindsuccessor++;
  find_route (x, wrap (mkref (this), &loctabroute::find_successor_cb, x, cb));
}

void
loctabroute::find_successor (const chordID &x, const bigint projkey, cbroute_t cb)
{
  nfindsuccessor++;
  find_route (x, projkey, wrap (mkref (this), &loctabroute::find_successor_cb, x, cb));
}

void
loctabroute::find_route (const chordID &x, cbroute_t cb) 
{
  route_iterator *ri = produce_iterator_ptr (x);
  ri->first_hop (wrap (this, &loctabroute::find_route_hop_cb, cb, ri), NULL);
}

void
loctabroute::find_route (const chordID &x, const bigint projkey, cbroute_t cb) 
{
  route_iterator *ri = produce_iterator_ptr (x, projkey);
  ri->first_hop (wrap (this, &loctabroute::find_route_hop_cb, cb, ri), NULL);
}

void
loctabroute::find_successor_cb (chordID x, cbroute_t cb, vec<chord_node> s, 
			       route search_path, chordstat status)
{
  if (status != CHORD_OK) {
    warnx << "find_successor_cb: find successor of " 
	  << x << " failed: " << status << "\n";
  } else {
    nhops += search_path.size ();
    if (search_path.size () > nmaxhops)
      nmaxhops = search_path.size ();
  }
  cb (s, search_path, status);
}

void
loctabroute::find_route_hop_cb (cbroute_t cb, route_iterator *ri, bool done)
{
  if (done) {
    cb (ri->successors (), ri->path (), ri->status ());
    delete ri;
  } else {
    ri->next_hop ();
  }
}

void
loctabroute::find_succlist_hop_cb (cbroute_t cb, route_iterator *ri, u_long m,
				  bool done)
{
  static bool shave = true;
  static bool initialized = false;
  if (!initialized) {
    int x = 1;
    assert (Configurator::only ().get_int ("chord.find_succlist_shaving", x));
    shave = (x == 1);
    initialized = true;
  }

  vec<chord_node> cs = ri->successors ();
  if (done) {
    cb (cs, ri->path (), ri->status ());
    delete ri;
    return;
  }
  if (shave) {
    size_t left = 0;
    if (cs.size () < m)
      left = cs.size ();
    else
      left = cs.size () - m;
    for (size_t i = 1; i < left; i++) {
      if (betweenrightincl (cs[i-1].x, cs[i].x, ri->key ())) {
	cs.popn_front (i);
	cb (cs, ri->path (), ri->status ());
	delete ri;
	return;
      }
    }
  }

  ri->next_hop ();
}

ptr<location> 
loctabroute::closestpred (const chordID &x, const vec<chordID> &f) 
{
  // Not sure what to do here.
  return NULL;
}

ptr<location> 
loctabroute::my_pred () const
{
  return locations->closestpredloc (my_ID());
}
ptr<location> 
loctabroute::my_succ () const
{
  return locations->closestsuccloc (incID (my_ID()));
}
ptr<location>
loctabroute::my_pred (bigint projkey)
{
  chordID first = my_ID();
  ptr<location> loc = my_pred();
  do {
    if ( !blacklist::only().deny_project(loc->address(), projkey) )
      return loc;
    loc = locations->prev_loc_circular (loc->id());
  } while(loc->id() != first);

  return NULL;
}
ptr<location>
loctabroute::my_succ (bigint projkey)
{
  chordID first = my_ID();
  ptr<location> loc = my_succ();
  do {
    if ( !blacklist::only().deny_project(loc->address(), projkey) )
      return loc;
    loc = locations->next_loc_circular (loc->id());
  } while(loc->id() != first);

  return NULL;
}

