/*
 *  Copyright (C) 2005  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef PTR_BLOCK_H
#define PTR_BLOCK_H

#include "stdinc.h"
#include "ns.h"

/* 
 * This class contains a list of pointers and a label for each pointer.
 * Memberlist subclasses this class.
 */

class ptr_block {

 public:
  ptr_block();
  ptr_block(const ptr_block &original);
  virtual ~ptr_block();

  virtual void read_block (bigint ptr_block_addr, callback<void, log_status>::ref cb);
  virtual bool add_entry (bigint pointer, str label);
  virtual bigint store (ptr<sfspriv> sk, callback<void, ns_stat, bigint>::ref cb);
  
  virtual bigint get_ptr (unsigned i) const;
  virtual str get_label (unsigned i) const;
  virtual unsigned nentries () const;
  virtual long get_blk_revision () const;
  virtual void print (str prefix) const;


 protected:
  vec<bigint> _pointers;
  vec<str> _labels;
  u_int64_t _fetcht;
  long _revision;
  
  virtual void got_ptr_block (callback<void, log_status>::ref cb,
			      ptr<keyhash_payload> p);
  virtual str serialize();
  virtual bool parse (xin *in);
  virtual void sort();
};

#endif /* PTR_BLOCK_H */
