#ifndef __PMAINT_H__
#define __PMAINT_H__

#include <chord.h>
#include <chord_prot.h>

#include <dhash_common.h>
#include <dhash.h>
#include <dhashcli.h>
#include <dhash_blockstore.h>
#include <location.h>
#include <ihash.h>
#include <id_utils.h>

typedef callback<void, ref<dbrec> >::ref delete_t;

class pmaint {

public:
  pmaint (dhashcli *cli, ptr<vnode> host_node, ptr<dhash_blockstore> _storage, 
	  delete_t delete_helper);

  void stop ();

  enum { PRTTMSHORT = 1, PRTTMLONG = 60, MAX_PENDING = 20};
  enum { PMAINT_HANDOFF_ERROR = 0, PMAINT_HANDOFF_NOTPRESENT = 1, 
	 PMAINT_HANDOFF_PRESENT = -1};

private:
  struct present_wrap {
    ihash_entry <present_wrap> link;
    bigint id;
    u_int count;
    present_wrap(bigint _id, u_int _count) { id = _id; count = _count; }
  };

  //helpers from parent class
  dhashcli *cli;
  ptr<vnode> host_node;
  ptr<dhash_blockstore> storage;
  delete_t delete_helper;

  bigint pmaint_offer_left;
  int pmaint_offer_next_succ;

  // count equals the number of times each key has been offered.
  //vec<u_int> pmaint_present_count; 
  ihash<bigint, present_wrap, &present_wrap::id, &present_wrap::link, hashID> pmaint_present_count;

  bigint pmaint_offer_right;
  bool pmaint_searching;
  bigint pmaint_next_key;

  vec<chord_node> pmaint_succs;

  timecb_t *active_cb;

  u_int get_present_count(bigint k);
  void inc_present_count(bigint k);
  void dec_present_count(bigint k);
  void clear_present_count();

  void pmaint_next ();
  void pmaint_lookup (chordID key, dhash_stat err, vec<chord_node> sl, route r);
  void pmaint_offer ();
  void pmaint_offer_cb (chord_node dst, vec<bigint> keys, ref<dhash_offer_res> res, 
			clnt_stat err);
  void pmaint_handoff (chord_node dst, bigint key, cbi cb);
  void pmaint_handoff_cb (bigint key, cbi cb, dhash_stat err, bool present);
  void handed_off_cb (chord_node dst,
		      vec<bigint> keys,
		     ref<dhash_offer_res> res,
		     int key_number,
		     int status);
  static bigint db_next (ptr<dhash_blockstore> db, bigint a);
  static vec<bigint> get_keys (ptr<dhash_blockstore> db, bigint a, bigint b, u_int maxcount);
};
#endif

