#ifndef _BLACKLIST_H_
#define _BLACKLIST_H_
/*
 *
 * Copyright (C) 2004 Alexander Yip (yipal@lcs.mit.edu)
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include "chord_types.h"
#include "id_utils.h"
#include <ihash.h>

str name_port2str(str name, int port);

struct project_entry
{
  ihash_entry <project_entry> link;
  bigint id;
  project_entry(bigint projkey) { id = projkey; }
};
typedef ihash<bigint, project_entry, &project_entry::id, 
  &project_entry::link, hashID> node_blacklist;

struct node_entry
{
  str name_port;
  ihash_entry<node_entry> link;

  int version;
  node_blacklist blist;
  node_entry(str name, int port) {
    name_port = name_port2str(name, port);
    version = 0;
  }
};

class blacklist {
 protected:
  str blacklist_dir;
  ihash<const str, node_entry, &node_entry::name_port,
    &node_entry::link> node_list;

  void init(str dirname, str name, int port);

 public:
  static blacklist &only ();
  int set_dir(str dirname);
  bool deny_project(str name, int port, bigint projkey);
  bool deny_project(const net_address &addr, bigint projkey);
  void append_blacklist(str name, int port, vec<bigint> &list);
  void append_blacklist(const net_address &addr, vec<bigint> &list);
  void read_all();
  int getversion(str name, int port);

  blacklist();
  ~blacklist();
};

#endif /* _BLACKLIST_H_ */
