
#include <fcntl.h>
#include "stdinc.h"
#include "async.h"
#include "str.h"
#include "penviron.h"
#include "stats.h"

past_stats::past_stats (int argc, char **argv) :
  _projid(0), _memberid(0), _logid(0), _server(""),
  _connected(false), _branched(false),
  _dht_bytes(0), _replica_bytes(0), _snapshot_bytes(0), _sent_bytes(0)
{
  int r = gettimeofday (&_time, NULL);
  if (r < 0) {
    _time.tv_sec = 0;
    _time.tv_usec = 0;
  }

  int msg_idx = -1;
  for(int i=0; i<argc; i++) {
    if (!strcmp(argv[i], "-m"))
      msg_idx = i+1;

    if (i == msg_idx && strlen(argv[i]) > 32 )
      _args.push_back (str (argv[i], 32));
    else
      _args.push_back (argv[i]);
  }
}

void
past_stats::set_projid (bigint projid)
{ _projid = projid; }

void
past_stats::set_memberid (bigint memberid) 
{ _memberid = memberid; }

void
past_stats:: set_logid (bigint logid) 
{ _logid = logid; }

void
past_stats::set_server_ip (str server)
{ 
  _server = (server && (server != "")) ? server : str("none");
}
  
void
past_stats::set_connected (bool connected) 
{ _connected = connected; }

void
past_stats::set_branched (bool branched) 
{ _branched = branched; }

void
past_stats::add_dht_bytes (unsigned new_bytes) 
{ _dht_bytes += new_bytes; }

void 
past_stats::add_replica_bytes (unsigned new_bytes) 
{ _replica_bytes += new_bytes; }

void
past_stats::add_snapshot_bytes (unsigned new_bytes)
{ _snapshot_bytes += new_bytes; }

void 
past_stats::add_sent_bytes (unsigned new_bytes)
{ _sent_bytes += new_bytes; }

void 
past_stats::append_to_log () 
{
  int fd = -1;
  unsigned max_trials = 3;
  str logfile = pget_statlog ();

  for (unsigned trials=0; trials<max_trials; trials++) {
    fd = open (logfile.cstr(), O_WRONLY | O_APPEND | O_CREAT | O_EXLOCK | O_NONBLOCK, 0644);
    if (fd >= 0 || (fd < 0 && errno != EWOULDBLOCK))
      break;
    sleep (1);
  }

  if (fd < 0) {
    warn << "Could not record stats, error " << errno
	 << " while opening stats log: " << logfile << "\n"; 
    return;
  }

  str s = gen_log_line ();
  size_t sent = 0;
  while (sent < s.len()) {
    int r = write (fd, s.cstr() + sent, s.len()-sent);
    if (r < 0) {
      warn << "Could not record stats, error " << errno 
	   << " while appending to stat log: " << logfile << "\n";
      break;
    }
    sent += r;
  }

  flock (fd, LOCK_UN);
  close (fd);
  return;
}
  
str 
past_stats::gen_log_line ()
{
  assert (_time.tv_usec < 1000000);
  char usecs[7];
  usecs[6] = 0;
  sprintf(usecs, "%06ld", _time.tv_usec);

  strbuf buf = strbuf() 
    << _time.tv_sec << "." << usecs << " "
    << armor64A(_projid) << " "
    << armor64A(_memberid) << " " 
    << armor64A(_logid) << " "
    << _server << " "
    << _connected << " "  << _branched << " "
    << _dht_bytes << " " << _sent_bytes << " "
    << _replica_bytes << " " << _snapshot_bytes << " ";

  for(unsigned i=0; i<_args.size(); i++) {
    buf << _args[i];
    if (i < _args.size()-1)
      buf << " ";
  }
  buf << "\n";

  return str(buf);
}

str
past_stats::read_log ()
{
  return file2str (pget_statlog ());
}

void
past_stats::truncate_log ()
{
  str logfile = pget_statlog ();
  int fd = open (logfile.cstr(), O_TRUNC | O_EXLOCK);
  if (fd >=0) {
    flock (fd, LOCK_UN);
    close (fd);
  }
}
